{-# OPTIONS -fglasgow-exts #-}

module Examples where

import Maybe
import List

import Data.Generics
import Traversal
import Infty
import Abbr

--------------------------------------------------------------------------------
-- The famous company examples
--------------------------------------------------------------------------------

data Company  = C [Dept]               deriving (Eq, Show, Typeable, Data)
data Dept     = D Name Manager [Unt]   deriving (Eq, Show, Typeable, Data)
data Unt      = PU Employee | DU Dept  deriving (Eq, Show, Typeable, Data)
data Employee = E Person Salary        deriving (Eq, Show, Typeable, Data)
data Person   = P Name Address         deriving (Eq, Show, Typeable, Data)
data Salary   = S Float                deriving (Eq, Show, Typeable, Data, Ord)
type Manager  = Employee
type Name     = String
type Address  = String

-- An illustrative company
genCom :: Company
genCom = C [D "Research" joe  [PU mike, PU kate],
            D "Strategy" mary []]

joe, mike, kate, mary :: Employee
joe  = E (P "Joe"  "Oregon") (S 8000)
mike = E (P "Mike" "Boston") (S 1000)
kate = E (P "Kate" "San Diego") (S 2000)
mary = E (P "Mary" "Washington") (S 100000)

incS :: Float -> Salary -> Salary
incS k (S s) = S (s * (1+k))

--- Increase salary

-- increase everyone's salary by 10%, topdown.
-- transformer, non-contextual, full, topdown, left-to-right
f1 :: Data t => Float -> t -> Maybe t
f1 k = traverse Trans NoCtx Full FromTop FromLeft (always $ incS k)

-- Same function, using syntactic sugars
increase :: Data t => Float -> t -> Maybe t
increase k = transformB (always $ incS k)

-- If we construct a visit with "always", and the coverage is full, it should
-- never return "Nothing". So it's safe to apply "fromJust"
ef1 = fromJust (f1 0.1 genCom)

-- increase everyone's salary by 10%, bottomup.
-- transformatin, non-contextual, full, bottomup, right-to-left
f1' :: Data t => Float -> t -> Maybe t
f1' k = traverse Trans NoCtx Full FromBottom FromRight (always $ incS k)

ef1' = fromJust (f1' 0.1 genCom)

-- These "once" traversals can really return "Nothing" even if the visits always
-- succeed, because they might not apply, in which case a "Nothing" should be
-- signaled.

-- increase the first person's salary by 10%, topdown, favoring left.
-- transformer, non-contextual, once, topdown, left-to-right
f3 :: Data t => Float -> t -> Maybe t
f3 k = traverse Trans NoCtx Once FromTop FromLeft (always $ incS k)

ef3 = f3 0.1 genCom

-- increase the first person's salary by 10%, topdown, favoring right.
-- transformer, non-contextual, once, topdown, right-to-left
f3' :: Data t => Float -> t -> Maybe t
f3' k = traverse Trans NoCtx Once FromTop FromRight (always $ incS k)

ef3' = f3' 0.1 genCom


--- Increase salary with a limited budget
incSbud :: Float -> Float -> Salary -> (Float, Salary)
incSbud k c (S s) = let i = min (s*k) c in (c-i,S (s+i))

-- increase everyone's salary by 10%, with a budget of $1000, favoring left and
-- top.
-- accumulative transformer, non-contextual, full, topdown, left-to-right
f5 :: Data t => Float -> Float -> t -> Maybe (Float,t)
f5 cap k = traverse AccTrans NoCtx Full FromTop FromLeft 
                    (always $ incSbud k) cap

ef5 = snd (fromJust (f5 1000 0.1 genCom))

-- increase everyone's salary by 10%, with a budget of $1000, favoring right
-- and bottom.
-- accumulative transformer, non-contextual, full, bottomup, right-to-left
f5' :: Data t => Float -> Float -> t -> Maybe (Float,t)
f5' cap k = traverse AccTrans NoCtx Full FromBottom FromRight 
                     (always $ incSbud k) cap

ef5' = snd (fromJust (f5' 1000 0.1 genCom))

-- Increase salary, collecting the amount being increased.
colS :: Float -> Float -> Salary -> (Float, Salary)
colS k a (S s) = (a+k*s, S (s*(1+k)))

-- increase everyone's salary by 10%, collecting the amount being increased
-- accumulative transformer, non-contextual, full, topdown, left-to-right
g1 :: Data t => Float -> t -> Maybe (Float,t)
g1 k = traverse AccTrans NoCtx Full FromTop FromLeft 
                (always (colS k)) 0

-- syntactic sugar
incBill :: Float -> Company -> Maybe (Float,Company)
incBill k = acctrans (always (colS k)) 0

eg1 = fromJust (g1 0.1 genCom)

-- increase the first person's salary by 10%, topdown, favoring left, collecting
-- the amount being increased.
-- accumulative transformer, non-contextual, once, topdown, left-to-right
g2 :: Data t => Float -> t -> Maybe (Float,t)
g2 k = traverse AccTrans NoCtx Once FromTop FromLeft 
                (always (colS k)) 0

eg2 = g2 0.1 genCom

-- increase the first person's salary by 10%, topdown, favoring right,
-- collecting the amount being increased.
-- accumulative transformer, non-contextual, once, topdown, right-to-left
g2' :: Data t => Float -> t -> Maybe (Float,t)
g2' k = traverse AccTrans NoCtx Once FromTop FromRight 
                 (always (colS k)) 0

eg2' = g2' 0.1 genCom

isDept n (D n' _ _) = n==n'

-- "local transformation". see \cite{vis04.strategoxt}
-- increase the salary of everyone in the research department by 10%.
-- transformer, non-contextual, stop, topdown, left-to-right
g4 :: Data t => String -> Float -> t -> Maybe t
g4 d k = traverse Trans NoCtx Stop FromTop FromLeft 
                  (f1 k `mwhenever` isDept d)

incOne :: Data t => String -> Float -> t -> Maybe t
incOne d k = stopTransform (f1 k `mwhenever` isDept d)

eg4 = fromJust (g4 "Research" 0.1 genCom)

-- increase the salary of everyone in the research department by 10%, returning
-- the amount being increased.
-- accumulative transformer, non-contextual, stop, topdown, left-to-right
g5 :: Data t => String -> Float -> t -> Maybe (Float,t)
g5 d k = traverse AccTrans NoCtx Stop FromTop FromLeft 
                  (f `mwhenever` \_ (D n _ _) -> n==d) 0
   where f a d = do (a',t) <- g1 k d
                    return (a+a',t)

eg5 = fromJust (g5 "Research" 0.1 genCom)

-- increase the salary of everyone in the research department by 10%, using
-- context.
-- transformer, contextual, full, topdown left-to-right
h1 :: Data t => Float -> String -> t -> Maybe t
h1 k d = traverse Trans Ctx Full FromTop FromLeft
           (mk (\c (D n _ _) -> c || n==d))
           (cond (\c s -> c) (\c s -> incS k s) (\c s -> s))
           False

eh1 = fromJust (h1 0.1 "Research" genCom)

-- increase rate depends on the department.
-- transformer, contextual, full, topdown left-to-right
incV :: Float -> Company -> Maybe Company
incV k = traverse Trans Ctx Full FromTop FromLeft 
                  (mk (\c d -> lookupRate d))
                  (always incS)
                  k

lookupRate :: Dept -> Float
lookupRate = const 0.1

-- find the smallest salary in the company
-- accumulator, non-contextual, full, topdown, left-to-right
h2 :: Data t => t -> Maybe FFloat
h2 t = traverse Accum NoCtx Full FromTop FromLeft 
                (always $ \a (S s) -> min a (F s)) Infty t

eh2 = fromJust (h2 genCom)

-- replace everyone's salary with a fixed number
-- transformer, non-contextual, full, topdown, left-to-right
h3 :: Data t => Float -> t -> Maybe t
h3 f = traverse Trans NoCtx Full FromTop FromLeft (always $ \(S _) -> S f)

eh3 = fromJust (h3 600 genCom)

-- replace everyone's salary with the lowest salary
-- accumulative transformer, non-contextual, full, topdown, left-to-right

-- there is a trick that relies on lazy evaluation. the whole procedure takes
-- only one pass. the transformation is evaluated on demand, when the
-- accumulation is already done.
h4 :: Data t => t -> Maybe (FFloat,t)
h4 t = traverse AccTrans NoCtx Full FromTop FromLeft 
                (always v) Infty t
    where v a (S s) = (min a (F s), S f)
              where Just (F f,_) = h4 t

eh4 = snd (fromJust (h4 genCom))

-- collect the sum of salaries
-- accumulator, non-contextual, full, topdown, left-to-right
h5 :: Data t => t -> Float
h5 t = fromJust $ 
       traverse Accum NoCtx Full FromTop FromLeft 
                (always col) 0 t
    where col a (S s) = a + s

-- syntactic sugar
bill :: Company -> Maybe Float
bill = accumulate (always col) 0
    where col a (S s) = a + s

eh5 = h5 genCom

-- determine the new salary based on the salary of the whole company
-- endless possibilities!
incSmt :: Data t => ([Float] -> [Float]) -> t -> ([Float],t)
incSmt scheme t = fromJust (acctrans (always v) [] t)
    where v a (S s) = (a++[s], S (new!!length a))
          new = scheme (fst (incSmt scheme t))

eh6 = snd (incSmt (socialism 1000 0.1) genCom)
eh6' = snd (incSmt (capitalism 10000 0.1) genCom)
eh6'' = snd (incSmt (communism 2000) genCom)

-- Three schemes.
-- increase everyone's salary by 10%, with a limited budget, favoring the poor.
socialism :: Float -> Float -> [Float] -> [Float]
socialism cap k xs = xs3
    where (idx,xs1) = unzip (sortBy (\x y -> compare (snd x) (snd y)) 
                                    (zip [1..] xs))
          (_,xs2) = foldl (\(c,xs) s -> let i = min (s*k) c 
                                        in (c-i,xs++[s+i])) 
                          (cap,[]) xs1
          (_,xs3) = unzip (sortBy (\x y -> compare (fst x) (fst y)) 
                                  (zip idx xs2))

capitalism :: Float -> Float -> [Float] -> [Float]
capitalism bud k ys = ys3
     where (ys1,xs)  = ixSort ys [1..]
           (_  ,ys2) = foldr f (bud,[]) ys1
           (_  ,ys3) = ixSort xs ys2
           f s (b,ys) = let i = min (s*k) b
                        in (b-i,(s+i):ys)

ixSort :: Ord a => [a] -> [b] -> ([a],[b])
ixSort xs ys  =
  unzip $ sortBy (\(x,_) (y,_)->compare x y) $ zip xs ys

-- average everyone's salary, with a total increase.
communism :: Float -> [Float] -> [Float]
communism cap xs = replicate (length xs) 
                              ((cap+sum xs) / (fromIntegral (length xs)))

--------------------------------------------------------------------------------
-- Lambda calculus examples
--------------------------------------------------------------------------------

data Var = V String
         deriving (Typeable, Data)

data Lam = Var Var
         | App Lam Lam
         | Abs String Lam
         deriving (Typeable, Data)

instance Show Lam where
    show (Var (V s)) = s
    show (App e1 e2) = show e1 ++ " " ++ show e2
    show (Abs v e) = "(\\" ++ v ++ "." ++ show e ++ ")"

vf = Var (V "f")
vx = Var (V "x")
vy = Var (V "y")
vu = Var (V "u")
vz = Var (V "z")

e1 = vf
e2 = App vf vx
e3 = App (Abs "x" (Abs "y" (App vx vy))) (Abs "x" vx)
e4 = App (App (Abs "x" (Abs "y" vx)) vu) vz
e5 = (Abs "x" (App vx vx))
e6 = App e5 e5
e7 = App (Abs "x" vy) e6
e8 = App (Abs "x" (App vx (Abs "x" vx))) vz

-- collect all variables in a lambda term
-- accumulator, non-contextual, full, topdown, left-to-right
vars :: Data t => t -> [String]
vars e = fromJust $
         traverse Accum NoCtx Full FromTop FromLeft 
                  (always f) [] e
    where f a (V v) = v:a

-- collect all FREE variables in a lambda term
-- accumulation, contextual, full, topdown, left-to-right
-- context: list of bound variables
-- accumulator: list of free variables
freevars :: Data t => t -> [String]
freevars e = fromJust $ 
             traverse Accum Ctx Full FromTop FromLeft (mk u) 
                      (always f) [] [] e
    where u bv (Abs v e) = v:bv
          u bv _ = bv
          f bv fv (V v) | v `elem` bv = fv
                        | otherwise = v:fv

-- handling beta redex
isRedex (App (Abs _ _) _) = True
isRedex _ = False

reduce (App (Abs v e) d) = 
    fromJust $ traverse Trans Ctx Full FromBottom FromLeft (mk upd) 
                        (always $ repl v d) [] e
reduce e = e

upd :: [Name] -> Lam -> [Name]
upd bv (Abs v _) = v:bv
upd bv _ = bv

repl :: Name -> Lam -> [Name] -> Lam -> Lam
repl v d bv e@(Var (V v')) | v'==v && notElem v bv = d 
repl _ _ _ e = e

-- applicative order evaluation
app_eval :: Data t => t -> t
app_eval = innermost Trans NoCtx FromLeft (reduce `whenever` isRedex)

-- normal order evaluation
norm_eval :: Data t => t -> t
norm_eval = outermost Trans NoCtx FromLeft (reduce `whenever` isRedex)

-- e3 = (\x.\y.x y) (\x.x) is an example to show that innermost and bottomup do
-- not work the same.
-- traverse Trans NoCtx Full FromBottom FromLeft (always reduce) e3

